import datetime
from typing import Optional

from cpl_core.database.context import DatabaseContextABC
from cpl_core.type import T, R
from cpl_query.extension import List

from bot_core.logging.database_logger import DatabaseLogger
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.model.user import User


class UserRepositoryService(UserRepositoryABC):
    def __init__(
        self,
        logger: DatabaseLogger,
        db_context: DatabaseContextABC,
        servers: ServerRepositoryABC,
    ):
        self._logger = logger
        self._context = db_context

        self._servers = servers

        UserRepositoryABC.__init__(self)

    def _from_result(self, result: tuple):
        return User(
            result[1],
            result[2],
            result[3],
            result[4],
            result[5],
            self._servers.get_server_by_id(result[6]),
            result[7],
            result[8],
            id=result[0],
        )

    def get_users(self) -> List[User]:
        self._logger.trace(__name__, f"Send SQL command: {User.get_select_all_string()}")
        return List(
            User, [self._from_result(user) for user in self._context.select(User.get_select_all_string())]
        ).for_each(lambda x: self._set_user_activity(x))

    def get_user_by_id(self, id: int) -> User:
        self._logger.trace(__name__, f"Send SQL command: {User.get_select_by_id_string(id)}")
        result = self._context.select(User.get_select_by_id_string(id))[0]

        return self._from_result(result)

    def find_user_by_id(self, id: int) -> Optional[User]:
        self._logger.trace(__name__, f"Send SQL command: {User.get_select_by_id_string(id)}")
        result = self._context.select(User.get_select_by_id_string(id))
        if result is None or len(result) == 0:
            return None

        return self._from_result(result[0])

    def get_users_by_discord_id(self, discord_id: int) -> List[User]:
        self._logger.trace(
            __name__,
            f"Send SQL command: {User.get_select_by_discord_id_string(discord_id)}",
        )
        return List(
            User,
            [
                self._from_result(user)
                for user in self._context.select(User.get_select_by_discord_id_string(discord_id))
            ],
        )

    def get_users_by_server_id(self, server_id: int) -> List[User]:
        self._logger.trace(
            __name__,
            f"Send SQL command: {User.get_select_by_server_id_string(server_id)}",
        )
        return List(
            User,
            [self._from_result(user) for user in self._context.select(User.get_select_by_server_id_string(server_id))],
        )

    def get_users_with_activity_by_server_id(self, server_id: int) -> List[User]:
        self._logger.trace(
            __name__,
            f"Send SQL command: {User.get_select_by_server_id_string(server_id)}",
        )
        return List(
            User,
            [self._from_result(user) for user in self._context.select(User.get_select_by_server_id_string(server_id))],
        ).for_each(lambda x: self._set_user_activity(x))

    def get_user_by_discord_id_and_server_id(self, discord_id: int, server_id: int) -> User:
        self._logger.trace(
            __name__,
            f"Send SQL command: {User.get_select_by_discord_id_and_server_id_string(discord_id, server_id)}",
        )

        return self._from_result(
            self._context.select(User.get_select_by_discord_id_and_server_id_string(discord_id, server_id))[0]
        )

    def find_user_by_discord_id_and_server_id(self, discord_id: int, server_id: int) -> Optional[User]:
        self._logger.trace(
            __name__,
            f"Send SQL command: {User.get_select_by_discord_id_and_server_id_string(discord_id, server_id)}",
        )
        result = self._context.select(User.get_select_by_discord_id_and_server_id_string(discord_id, server_id))
        if result is None or len(result) == 0:
            return None

        return self._from_result(result[0])

    def add_user(self, user: User):
        self._logger.trace(__name__, f"Send SQL command: {user.insert_string}")
        self._context.cursor.execute(user.insert_string)

    def update_user(self, user: User):
        self._logger.trace(__name__, f"Send SQL command: {user.udpate_string}")
        self._context.cursor.execute(user.udpate_string)

    def delete_user(self, user: User):
        self._logger.trace(__name__, f"Send SQL command: {user.delete_string}")
        self._context.cursor.execute(user.delete_string)

    def _set_user_activity(self, user):
        days = (datetime.date.today() - (datetime.date.today() - datetime.timedelta(days=7))).days
        query = f"""
            SELECT (
                (
                    SELECT Count(UserGotAchievements.CreatedAt)
                    FROM UserGotAchievements
                    INNER JOIN Achievements ON UserGotAchievements.AchievementId = Achievements.Id
                    INNER JOIN Users ON UserGotAchievements.UserId = Users.UserId
                    WHERE Users.ServerId = {user.server.id}
                    AND Users.UserId = {user.id}
                    AND UserGotAchievements.CreatedAt >= "{datetime.date.today() - datetime.timedelta(days=7)}"
                ) +
                (
                    SELECT SUM(
                        UserMessageCountPerHour.XPCount / (
                            SELECT XpPerMessage
                            FROM CFG_Server
                            WHERE ServerId = {user.server.id}
                        )
                    )
                    FROM UserMessageCountPerHour
                    INNER JOIN Users ON UserMessageCountPerHour.UserId = Users.UserId
                    WHERE Users.ServerId = {user.server.id}
                    AND Users.UserId = {user.id}
                    AND UserMessageCountPerHour.CreatedAt >= "{datetime.date.today() - datetime.timedelta(days=7)}"
                ) +
                (
                    SELECT Count(UserJoinedVoiceChannel.CreatedAt)
                    FROM UserJoinedVoiceChannel
                    INNER JOIN Users ON UserJoinedVoiceChannel.UserId = Users.UserId
                    WHERE Users.ServerId = {user.server.id}
                    AND Users.UserId = {user.id}
                    AND UserJoinedVoiceChannel.CreatedAt >= "{datetime.date.today() - datetime.timedelta(days=7)}"
                ) +
                (
                    SELECT IFNULL(ROUND(SUM(TIME_TO_SEC(
                        TIMEDIFF(UserJoinedVoiceChannel.LeavedOn, UserJoinedVoiceChannel.JoinedOn)
                    ) / 3600), 2), 0)
                    FROM UserJoinedVoiceChannel
                    INNER JOIN Users ON UserJoinedVoiceChannel.UserId = Users.UserId
                    WHERE Users.ServerId = {user.server.id}
                    AND Users.UserId = {user.id}
                    AND UserJoinedVoiceChannel.CreatedAt >= "{datetime.date.today() - datetime.timedelta(days=7)}"
                ) +
                (
                    SELECT Count(UserJoinedGameServer.CreatedAt)
                    FROM UserJoinedGameServer
                    INNER JOIN Users ON UserJoinedGameServer.UserId = Users.UserId
                    WHERE Users.ServerId = {user.server.id}
                    AND Users.UserId = {user.id}
                    AND UserJoinedGameServer.CreatedAt >= "{datetime.date.today() - datetime.timedelta(days=7)}"
                ) +
                (
                    SELECT IFNULL(ROUND(SUM(TIME_TO_SEC(
                        TIMEDIFF(UserJoinedGameServer.LeavedOn, UserJoinedGameServer.JoinedOn)
                    ) / 3600), 2), 0)
                    FROM UserJoinedGameServer
                    INNER JOIN Users ON UserJoinedGameServer.UserId = Users.UserId
                    WHERE Users.ServerId = {user.server.id}
                    AND Users.UserId = {user.id}
                    AND UserJoinedGameServer.CreatedAt >= "{datetime.date.today() - datetime.timedelta(days=7)}"
                ) -
                (
                    SELECT COUNT(UserWarnings.CreatedAt)
                    FROM UserWarnings
                    INNER JOIN Users ON UserWarnings.UserId = Users.UserId
                    WHERE Users.ServerId = {user.server.id}
                    AND Users.UserId = {user.id}
                    AND UserWarnings.CreatedAt >= "{datetime.date.today() - datetime.timedelta(days=7)}"
                )
           ) / {days} * 1000 as count;
        """
        user.activity = self._cast_query_result(query, int)

    def _cast_query_result(self, query: str, r_type: T) -> Optional[R]:
        results = self._context.select(query)
        if len(results) == 0 or len(results[0]) == 0:
            return None
        result = results[0][0]
        default = None
        if r_type is int or r_type is float:
            default = 0
        elif r_type is str:
            default = ""

        return r_type(result) if result is not None else default
